/*
 * File  : CTMS_Example_for_Raspberry.c:
 * version : 1.0
 * Date : 2020. 6. 16.
 * Author : TK Lee
 * Contact : expoeb2@diwell.com 
 * Corporation : Diwell Electronics
 *
 * description : 
 *   CTMS example code.  
 *   (using wiringPi Library)
 *
 * Further information:
 *   https://projects.drogon.net/raspberry-pi/wiringpi/
 *   https://www.diwellshop.com/
 *
 * license : free.
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY.
 ***********************************************************************
 * Note:
 *   Raspberry Pi 2 Model B v1.1 used for testing.
 *   This code has not been tested on other versions of the Raspberry. 
 ***********************************************************************
 * Revision history.
 * 1. 2020.6.16 : First version is released.
 ***********************************************************************
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <wiringPi.h>
#include <wiringSerial.h>
#include "desc.h"

#define MAX_BUFFER 	13		// RX max buffer
#define REQUEST_CYCLE 	100		// 100ms : Request cycle. Should be at least 100ms
#define TIMEOUT	20			// 20ms : Note. Do not modify this value!!!
#define TRANSMISSION_TIME	5	// 5ms : Note. Do not modify this value!!!

query_t telegram;			// Master query structure

uint8_t *ptr, com_result;
uint8_t RX_BUF[MAX_BUFFER];
uint16_t au16data[4];
uint16_t Request_interval = REQUEST_CYCLE-TIMEOUT-TRANSMISSION_TIME; // 100ms-20ms-5ms = 75ms
int32_t fd;
uint32_t Start_ms, request_count=0;
float RawData, CompData, SensData;

int main ()
{	
	// Never change the setting below.
   	telegram.u8id = 1;                               // Start : 1
   	telegram.u8fct = 3;                              // function code : 03  
   	telegram.u16RegAdd = 1;                          // start address : 1
	telegram.u16RegNo = 4;                           // number of elements to read : 4 
	telegram.au16reg = au16data;                     // pointer to a memory array in the raspberry 

	if ((fd = serialOpen ("/dev/ttyAMA0", 19200)) < 0)
	{
		fprintf (stderr, "Unable to open serial device: %s\n", strerror (errno)) ;
		return 1 ;
	}
	if (wiringPiSetup () == -1)
	{
		fprintf (stdout, "Unable to start wiringPi: %s\n", strerror (errno)) ;
		return 1 ;
	}
	
        delay(1000);                   			// waiting for Sensor init. (*min : 1sec)

	while(request_count++ < 100)  			// repeat 100 times.
  	{
		com_result = query(telegram);		// communication.

    		if(com_result == CRC_OK)
		{
			RawData = (float)au16data[0]*0.02-273.15; // DataValue1 : Skin Temp.(Raw Data)
			CompData = (float)au16data[1]*0.02-273.15;// DataValue2 : Skin Temp.
			//dummy = (float)au16data[2]*0.02-273.15; // DataValue3 : reserved(not used)
			SensData = (float)au16data[3]*0.02-273.15;// DataValue4 : Sensor package Temp.
			
			printf ("%4d- Skin Temp(Raw): %2.2f  Skin Temp: %2.2f Sensor Temp: %2.2f\n", request_count,RawData,CompData,SensData);
		}
		else if(com_result == CRC_ERROR)
		{
			printf("%4d- CRC Error. Check the line.\n", request_count);
		}
		else if(com_result == NO_REPLY)
		{
			printf("%4d- No reply. Check the line.\n", request_count);
		}
		
		Start_ms = millis();
		while( millis()-Start_ms < Request_interval); 		// Delay : next request
 	}
 	printf ("================Stopped================\n") ;
  	return 0 ;
}

int8_t query(query_t telegram)
{
	uint8_t au8Buffer[8], tx_count,rx_count=0,response_status = NO_REPLY;
	uint16_t crc16, *au16regs;

	au16regs = telegram.au16reg;
 	au8Buffer[ 0 ]  = telegram.u8id;
    	au8Buffer[ 1 ]  = telegram.u8fct;
    	au8Buffer[ 2 ]  = highByte(telegram.u16RegAdd );
    	au8Buffer[ 3 ]  = lowByte( telegram.u16RegAdd );
	au8Buffer[ 4 ]  = highByte(telegram.u16RegNo );
        au8Buffer[ 5 ]  = lowByte( telegram.u16RegNo );
	crc16 = CalcCRC16(au8Buffer, 6);
	au8Buffer[ 6 ]  = lowByte(crc16);   // reverse 
        au8Buffer[ 7 ]  = highByte(crc16);  // reverse

	fflush (stdout) ;			// flush buffer
	ptr = au8Buffer;
	
	for(tx_count=0; tx_count<8; tx_count++)
	{
		serialPutchar (fd, *ptr++) ; 	// transfer data
	}
	Start_ms = millis();
	// Note. Wait for transmission.
	while( millis()-Start_ms < TRANSMISSION_TIME);

	Start_ms = millis();
	while( millis()-Start_ms < TIMEOUT)	
	{
		while (serialDataAvail (fd))
		{ 
			RX_BUF[rx_count] = serialGetchar (fd);
	
			switch(rx_count)
			{
			case 0:
				if(RX_BUF[0] == 0x01){rx_count++;}
				break;
			case 1:
				if(RX_BUF[1] == 0x03){rx_count++;}
				else{rx_count=0;}
				break;
			case 2:
				if(RX_BUF[2] == 0x08){rx_count++;}
				else{rx_count=0;}
				break;
			case 3 ... 11:
				rx_count++;
				break;
			case 12:
				crc16 = CalcCRC16(RX_BUF, rx_count-1);			// CRC calculation.

				if(crc16 == (RX_BUF[rx_count-1]|RX_BUF[rx_count]<<8))	// CRC verification.
				{	
					au16regs[0] = (uint16_t)RX_BUF[3]<<8 |(uint16_t)RX_BUF[4];
					au16regs[1] = (uint16_t)RX_BUF[5]<<8 |(uint16_t)RX_BUF[6];
					au16regs[2] = (uint16_t)RX_BUF[7]<<8 |(uint16_t)RX_BUF[8];
					au16regs[3] = (uint16_t)RX_BUF[9]<<8 |(uint16_t)RX_BUF[10];
					response_status = CRC_OK;
				}
				else
				{
					response_status = CRC_ERROR;
				}
				rx_count=0;   
				break;
			default: rx_count=0; break;
	
			} // end switch(rx_count)
		} //end while (serialDataAvail (fd))
	}
	return response_status;
}

uint8_t highByte(uint16_t Data)
{
	return (Data>>8)&0xFF;
}

uint8_t lowByte(uint16_t Data)
{
	return Data&0xFF;
}

uint16_t CalcCRC16(uint8_t *cdata, uint16_t ilen)
{
    uint16_t AccumCRC16 = 0xffff;
    uint8_t i,j;
    for(j=0;j<ilen;j++)
    {
        i= (AccumCRC16 ^ *(cdata++) ) & 255;
        AccumCRC16 = ((AccumCRC16>>8) ^ TableCRC16[i]) & 0xffff;
    }
    return AccumCRC16;
}

